import json
import logging
import os
import re
import shutil
import sys
import tempfile
import time
import unittest
import warnings
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent.parent))

from pypdl import Pypdl


class TestPypdl(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestPypdl, self).__init__(*args, **kwargs)
        self.temp_dir = os.path.join(tempfile.gettempdir(), "pypdl_test")
        self.download_file_1MB = "https://7-zip.org/a/7z2409-src.tar.xz"
        self.no_head_support_url = "https://ash-speed.hetzner.com/100MB.bin"

        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)
        os.mkdir(self.temp_dir)

    def setUp(self):
        warnings.filterwarnings(
            "ignore", message="unclosed <socket.socket", category=ResourceWarning
        )
        warnings.filterwarnings(
            "ignore",
            message="unclosed transport",
            category=ResourceWarning,
        )

    def _assert_download(self, expected, success, filepath):
        self.assertEqual(success, expected, f"{expected - success} downloads failed")
        for path in filepath:
            self.assertTrue(os.path.exists(path))

    def test_single_segment_download(self):
        dl = Pypdl()
        url = self.download_file_1MB
        file_path = os.path.join(self.temp_dir)
        result = dl.start(url, file_path, display=False, multisegment=False)
        success = len(result)
        self.assertEqual(success, 1, "Download failed")

    def test_multi_segment_download(self):
        dl = Pypdl()
        url = self.download_file_1MB
        file_path = os.path.join(self.temp_dir, "test.dat")
        future = dl.start(url, file_path, display=False, block=False, speed_limit=0.5)
        time.sleep(2)
        self.assertTrue(os.path.exists(file_path + ".json"))
        success = len(future.result())
        self._assert_download(1, success, [file_path])

    def test_multiple_downloads(self):
        dl = Pypdl(max_concurrent=2)
        filepath = [os.path.join(self.temp_dir, f"file{i}.dat") for i in range(4)]
        tasks = [
            {
                "url": self.download_file_1MB,
                "file_path": filepath[0],
            },
            {
                "url": self.download_file_1MB,
                "file_path": filepath[1],
            },
            {
                "url": self.download_file_1MB,
                "file_path": filepath[2],
            },
            {
                "url": self.download_file_1MB,
                "file_path": filepath[3],
            },
        ]
        result = dl.start(tasks=tasks, block=True, display=False)
        success = len(result)
        self._assert_download(4, success, filepath)

    def test_allow_reuse(self):
        result = []
        dl = Pypdl(allow_reuse=True)
        url = self.download_file_1MB
        filepath = [os.path.join(self.temp_dir, f"file{i}.dat") for i in range(2)]

        res1 = dl.start(url, filepath[0], display=False)
        result.append(res1)
        res2 = dl.start(url, filepath[1], display=False)
        result.append(res2)
        success = len(result)
        dl.shutdown()
        self._assert_download(2, success, filepath)

    def test_speed_limit(self):
        dl = Pypdl()
        url = self.download_file_1MB  # file is 1.44MB, accurate time is 14s
        file_path = os.path.join(self.temp_dir, "test_1.dat")
        dl.start(url, file_path, display=False, speed_limit=0.1)
        self.assertTrue(10 <= int(dl.time_spent) <= 30, f"took: {dl.time_spent}s")

    def test_unblocked_download(self):
        dl = Pypdl(max_concurrent=2)
        filepath = [os.path.join(self.temp_dir, f"file{i}.dat") for i in range(2)]
        tasks = [
            {
                "url": self.download_file_1MB,
                "file_path": filepath[0],
            },
            {
                "url": self.download_file_1MB,
                "file_path": filepath[1],
            },
        ]
        future = dl.start(tasks=tasks, block=False, display=False)
        while not dl.completed:
            time.sleep(1)
        success = len(future.result())
        self._assert_download(2, success, filepath)

    def test_stop_restart_with_segements(self):
        dl = Pypdl(max_concurrent=2)
        url = self.download_file_1MB
        file_path = os.path.join(self.temp_dir, "test.dat")
        dl.start(
            url, file_path, block=False, display=False, speed_limit=0.15, segments=3
        )
        time.sleep(5)
        progress = dl.progress
        dl.stop()
        self.assertTrue(os.path.exists(file_path + ".0"))
        self.assertTrue(os.path.exists(file_path + ".1"))
        self.assertTrue(dl.is_idle)
        dl.start(url, file_path, display=False, speed_limit=0.01, block=False)
        time.sleep(3)
        self.assertTrue(dl.progress >= progress)
        dl.stop()
        self.assertTrue(dl.is_idle)
        res = dl.start(url, file_path, display=False)
        success = len(res)
        self._assert_download(1, success, [file_path])

    def test_logger(self):
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        log_file = os.path.join(self.temp_dir, "test.log")
        handler = logging.FileHandler(log_file, mode="a", delay=True)
        logger.addHandler(handler)

        dl = Pypdl(logger=logger)
        url = self.download_file_1MB
        file_path = os.path.join(self.temp_dir, "test.dat")
        dl.start(url, file_path, display=False, block=False)
        time.sleep(3)
        dl.shutdown()
        logger.removeHandler(handler)
        handler.close()
        self.assertTrue(os.path.exists(log_file))

    def test_header(self):
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        log_file = os.path.join(self.temp_dir, "header.log")
        handler = logging.FileHandler(log_file, mode="a", delay=True)
        logger.addHandler(handler)

        dl = Pypdl(allow_reuse=True, logger=logger)
        url = self.download_file_1MB
        file_path = os.path.join(self.temp_dir, "test.dat")
        dl.start(url, file_path, display=False, block=False, speed_limit=0.1)
        time.sleep(3)
        dl.stop()
        self.assertTrue(os.path.exists(log_file))
        with open(log_file, "r") as f:
            log_content = f.read()
        header_pattern = re.compile("Header acquired from HEAD request")
        self.assertTrue(
            header_pattern.search(log_content),
            "Unable to acquire header from HEAD request",
        )

        url = self.no_head_support_url
        file_path = os.path.join(self.temp_dir, "temp.csv")
        dl.start(url, file_path, display=False, block=False, speed_limit=0.1)
        time.sleep(10)

        dl.shutdown()
        logger.removeHandler(handler)
        handler.close()

        with open(log_file, "r") as f:
            log_content = f.read()
        header_pattern = re.compile("Header acquired from GET request")
        self.assertTrue(
            header_pattern.search(log_content),
            "Unable to acquire header from GET request",
        )

    def test_retries(self):
        mirrors = [self.download_file_1MB, "http://fake_website/file2"]
        file_path = os.path.join(self.temp_dir, "test.dat")
        dl = Pypdl()
        res = dl.start(
            "http://fake_website/file",
            file_path,
            display=False,
            retries=2,
            mirrors=mirrors,
        )
        success = len(res)
        self._assert_download(1, success, [file_path])

    def test_overwrite(self):
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        log_file = os.path.join(self.temp_dir, "overwrite.log")
        handler = logging.FileHandler(log_file, mode="a", delay=True)
        logger.addHandler(handler)

        dl = Pypdl(logger=logger, allow_reuse=True)
        url = self.download_file_1MB
        file_path = os.path.join(self.temp_dir, "test.dat")
        res = dl.start(url, file_path, display=False)
        success = len(res)
        self._assert_download(1, success, [file_path])
        res = dl.start(url, file_path, display=False, overwrite=False)
        success = len(res)
        self._assert_download(1, success, [file_path])

        with open(log_file, "r") as f:
            log_content = f.read()
        header_pattern = re.compile("File already exists, download completed")
        self.assertTrue(header_pattern.search(log_content), "overwrite not working")
        dl.shutdown()
        logger.removeHandler(handler)
        handler.close()

    def test_etag_validation(self):
        dl = Pypdl()
        url = self.download_file_1MB
        file_path = os.path.join(self.temp_dir, "test.dat")
        dl.start(url, file_path, display=False, block=False, speed_limit=0.1)
        time.sleep(5)
        progress = dl.progress
        dl.stop()

        with open(file_path + ".json", "r") as f:
            json_file = json.load(f)

        json_file["etag"] = "fake_etag"

        with open(file_path + ".json", "w") as f:
            json.dump(json_file, f)

        dl.start(
            url,
            file_path,
            display=False,
            etag_validation=False,
            speed_limit=0.01,
            block=False,
        )
        time.sleep(3)
        self.assertTrue(dl.progress >= progress)
        dl.shutdown()

    def test_failed(self):
        dl = Pypdl()
        url = "http://fake_website/file"
        file_path = os.path.join(self.temp_dir, "test.dat")
        dl.start(url, file_path, display=False)
        self.assertEqual(len(dl.failed), 1, "Failed downloads not found")


if __name__ == "__main__":
    unittest.main(verbosity=2)



import asyncio
from threading import Lock

from aiofiles import os

from .downloader import Multidown, Singledown
from .utils import (
    FileValidator,
    auto_cancel_gather,
    check_main_thread_exception,
    combine_files,
    create_segment_table,
)


class Consumer:
    def __init__(self, session, logger, _id):
        self._workers = []
        self._downloaded_size = 0
        self._size = 0
        self._success = []
        self._show_size = True
        self._lock = Lock()
        self._id = _id
        self._logger = logger
        self._session = session

    @property
    def size(self):
        if self._show_size:
            self._size = (
                sum(worker.curr for worker in self._workers) + self._downloaded_size
            )
        return self._size

    @property
    def success(self):
        with self._lock:
            return self._success.copy()

    def add_success(self, item):
        with self._lock:
            self._success.append(item)

    async def process_tasks(self, in_queue, out_queue):
        self._logger.debug("Consumer %s started", self._id)
        while True:
            task = await in_queue.get()
            if task is None:
                break
            self._logger.debug("Consumer %s received task", self._id)
            try:
                await self._download(task)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                check_main_thread_exception(e)
                self._logger.debug("Task %s failed", self._id)
                self._logger.exception(e)
                await out_queue.put([task[0]])

            # Clear workers after task completion
            self._workers.clear()
            self._show_size = True

            self._logger.debug("Consumer %s completed task", self._id)
        self._logger.debug("Consumer %s exited", self._id)

    async def _download(
        self,
        task,
    ):
        _id, task = task
        (
            url,
            file_path,
            multisegment,
            etag,
            size,
            segments,
            overwrite,
            speed_limit,
            etag_validation,
            kwargs,
        ) = task

        self._logger.debug("Download started %s", self._id)
        if not overwrite and await os.path.exists(file_path):
            self._logger.debug("File already exists, download completed")
            self.add_success((url, FileValidator(file_path)))
            self._downloaded_size += await os.path.getsize(file_path)
            return

        if multisegment:
            segment_table = await create_segment_table(
                url, file_path, segments, size, etag, etag_validation
            )
            await self._multi_segment(segment_table, file_path, speed_limit, **kwargs)
        else:
            await self._single_segment(url, file_path, speed_limit, **kwargs)

        self.add_success((url, FileValidator(file_path)))
        self._logger.debug("Download exited %s", self._id)

    async def _multi_segment(self, segment_table, file_path, speed_limit, **kwargs):
        tasks = set()
        segments = segment_table["segments"]
        speed_limit = speed_limit / segments
        self._logger.debug("Multi-Segment download started %s", self._id)
        for segment in range(segments):
            md = Multidown(self._session, speed_limit)
            self._workers.append(md)
            tasks.add(asyncio.create_task(md.worker(segment_table, segment, **kwargs)))

        await auto_cancel_gather(*tasks)
        await combine_files(file_path, segments)
        self._logger.debug("Downloaded all segments %s", self._id)
        self._show_size = False
        self._downloaded_size += await os.path.getsize(file_path)

    async def _single_segment(self, url, file_path, speed_limit, **kwargs):
        self._logger.debug("Single-Segment download started %s", self._id)
        sd = Singledown(self._session, speed_limit)
        self._workers.append(sd)
        await sd.worker(url, file_path, **kwargs)
        self._logger.debug("Downloaded single segment %s", self._id)
        self._show_size = False
        self._downloaded_size += await os.path.getsize(file_path)



import asyncio
import hashlib
import json
import logging
import sys
import time
from concurrent.futures import CancelledError, Executor, Future, ThreadPoolExecutor
from os import path
from threading import Event, Thread
from typing import Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import unquote, urlparse

from aiofiles import open as fopen
from aiofiles import os as aio_os

MEGABYTE = 1048576
BLOCKSIZE = 4096
BLOCKS = 1024
CHUNKSIZE = BLOCKSIZE * BLOCKS
MAX_FILENAME_LENGTH = 255


class MainThreadException(Exception):
    pass


class Size:
    def __init__(self, start: int, end: int) -> None:
        self.start = start
        self.end = end
        self.value = end - start + 1  # since range is inclusive[0-99 -> 100]

    def __repr__(self) -> str:
        return str(self.value)


class Task:
    def __init__(
        self,
        multisegment: bool,
        segments: int,
        tries: int,
        overwrite: bool,
        speed_limit: Union[float, int],
        etag_validation: bool,
        **kwargs,
    ):
        self.url = None
        self.file_path = None
        self.mirrors = None
        self.default_url = None
        self.multisegment = multisegment
        self.segments = segments
        self.tries = tries + 1
        self.overwrite = overwrite
        self.speed_limit = speed_limit
        self.etag_validation = etag_validation
        self.size = Size(0, 0)
        self.kwargs = kwargs if kwargs else {}

    def set(self, **kwargs) -> None:
        for key, value in kwargs.items():
            if key == "retries":
                key = "tries"
                value = value + 1

            if hasattr(self, key):
                setattr(self, key, value)
            else:
                self.kwargs[key] = value
        self.validate()

        if self.mirrors is not None and not isinstance(self.mirrors, list):
            self.mirrors = [self.mirrors]
        self.default_url = self.url

    def validate(self) -> None:
        if not (isinstance(self.url, str) or callable(self.url)):
            raise TypeError(
                f"url should be of type str or callable, got {type(self.url).__name__}"
            )
        if not (isinstance(self.file_path, str) or self.file_path is None):
            raise TypeError(
                f"file_path should be of type str or None, got {type(self.file_path).__name__}"
            )
        if not (
            isinstance(self.mirrors, (list, str))
            or self.mirrors is None
            or callable(self.mirrors)
        ):
            raise TypeError(
                f"mirrors should be of type str, callable, list or None, got {type(self.mirrors).__name__}"
            )
        if not isinstance(self.multisegment, bool):
            raise TypeError(
                f"multisegment should be of type bool, got {type(self.multisegment).__name__}"
            )
        if not isinstance(self.segments, int):
            raise TypeError(
                f"segments should be of type int, got {type(self.segments).__name__}"
            )
        if not isinstance(self.tries, int):
            raise TypeError(
                f"tries should be of type int, got {type(self.tries).__name__}"
            )
        if not isinstance(self.overwrite, bool):
            raise TypeError(
                f"overwrite should be of type bool, got {type(self.overwrite).__name__}"
            )
        if not isinstance(self.speed_limit, (float, int)):
            raise TypeError(
                f"speed_limit should be of type float or int, got {type(self.speed_limit).__name__}"
            )
        if not isinstance(self.etag_validation, bool):
            raise TypeError(
                f"etag_validation should be of type bool, got {type(self.etag_validation).__name__}"
            )

    def __repr__(self) -> str:
        return f"Task(url={self.url}, file_path={self.file_path}, tries={self.tries}, size={self.size})"


class TEventLoop:
    """A Threaded Eventloop"""

    def __init__(self):
        self.loop = asyncio.new_event_loop()
        self._thread = Thread(target=self._run, daemon=True)
        self._thread.start()

    def _run(self) -> None:
        self.loop.run_forever()
        self.loop.close()

    def get(self) -> asyncio.AbstractEventLoop:
        return self.loop

    def call_soon_threadsafe(self, func, *args) -> None:
        return self.loop.call_soon_threadsafe(func, *args)

    def has_running_tasks(self) -> bool:
        tasks = asyncio.all_tasks(self.loop)
        return any(not task.done() for task in tasks)

    def clear_wait(self) -> None:
        while self.has_running_tasks():
            time.sleep(0.1)

    def stop(self, *args) -> None:
        self.clear_wait()
        self.call_soon_threadsafe(self.loop.stop)
        self._thread.join()


class LoggingExecutor:
    """An Executor that logs exceptions."""

    def __init__(self, logger: logging.Logger, *args, **kwargs):
        self.executor = ThreadPoolExecutor(*args, **kwargs)
        self.logger = logger

    def submit(self, func: Callable, *args, **kwargs) -> Future:
        return self.executor.submit(self._wrap(func, *args, **kwargs))

    def shutdown(self) -> None:
        self.executor.shutdown()

    def _wrap(self, func: Callable, *args, **kwargs) -> Callable:
        def wrapper():
            try:
                return func(*args, **kwargs)
            except Exception as e:
                self.logger.exception(e)

        return wrapper


class FileValidator:
    """A class used to validate the integrity of the file."""

    def __init__(self, path: str):
        self.path = path

    def calculate_hash(self, algorithm: str, **kwargs) -> str:
        hash_obj = hashlib.new(algorithm, **kwargs)
        with open(self.path, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                hash_obj.update(chunk)
        return hash_obj.hexdigest()

    def validate_hash(self, correct_hash: str, algorithm: str, **kwargs) -> bool:
        file_hash = self.calculate_hash(algorithm, **kwargs)
        return file_hash == correct_hash


class AutoShutdownFuture:
    """A Future object wrapper that shuts down the eventloop and executor when the result is retrieved."""

    def __init__(self, future: Future, loop: TEventLoop, executor: Executor):
        self._future = future
        self._executor = executor
        self._loop = loop

    def result(
        self, timeout: Optional[float] = None
    ) -> Union[List[FileValidator], None]:
        result = self._future.result(timeout)
        self._loop.stop()
        self._executor.shutdown()
        return result


class EFuture:
    """A Future object wrapper that cancels the future and clears the eventloop when stopped."""

    def __init__(self, future: Future, loop: TEventLoop, interrupt: Event):
        self._future = future
        self._loop = loop
        self._interrupt = interrupt

    def result(
        self, timeout: Optional[float] = None
    ) -> Union[List[FileValidator], None]:
        try:
            while not self._future.done():
                time.sleep(1)
        except KeyboardInterrupt:
            self._stop()
            self._loop.stop()
            raise

        if self._future.done():
            return self._future.result(timeout)

    def _stop(self) -> None:
        self._loop.call_soon_threadsafe(self._future.cancel)
        try:
            self.result()
        except CancelledError:
            pass

        self._loop.clear_wait()
        self._interrupt.set()


class ScreenCleaner:
    """A context manager to clear the screen and hide cursor."""

    def __init__(self, display: bool, clear_terminal: bool):
        self.display = display
        self.clear_terminal = clear_terminal

    def clear(self) -> None:
        sys.stdout.write(2 * "\n")
        if self.clear_terminal:
            sys.stdout.write("\033c")  # Clear screen
        sys.stdout.write("\x1b[?25l")  # Hide cursor
        sys.stdout.flush()

    def __enter__(self):
        if self.display:
            self.clear()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.display:
            sys.stdout.write("\x1b[?25h")  # Show cursor
            sys.stdout.flush()


def to_mb(size_in_bytes: int) -> float:
    return size_in_bytes / MEGABYTE


def seconds_to_hms(sec: float) -> str:
    if sec == -1:
        return "99:59:59"
    time_struct = time.gmtime(sec)
    return time.strftime("%H:%M:%S", time_struct)


def cursor_up() -> None:
    sys.stdout.write("\x1b[1A" * 2)  # Move cursor up two lines
    sys.stdout.flush()


def check_main_thread_exception(e: Exception) -> None:
    if str(e) == "cannot schedule new futures after shutdown":
        raise MainThreadException from e


async def get_url(url: Union[str, Callable]) -> str:
    if callable(url):
        if asyncio.iscoroutinefunction(url):
            url = await url()
        else:
            url = url()

    if isinstance(url, str):
        return url
    raise TypeError(f"Function returned a non-string URL, got {type(url).__name__}")


async def auto_cancel_gather(*args, **kwargs) -> List:
    tasks = []
    for task in args:
        if isinstance(task, asyncio.Task):
            tasks.append(task)
        else:
            tasks.append(asyncio.create_task(task))
    try:
        return await asyncio.gather(*tasks, **kwargs)
    except Exception:
        for task in tasks:
            task.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)
        raise


async def get_filepath(url: str, headers: Dict[str, str], file_path: str) -> str:
    # Try to get filename from Content-Disposition header
    filename = None
    content_disposition = headers.get("Content-Disposition", "")

    # Check for filename* parameter (RFC 6266)
    if "filename*=" in content_disposition:
        try:
            start_pos = content_disposition.find("filename*=") + len("filename*=")
            end_pos = content_disposition.find(";", start_pos)
            encoded_filename = content_disposition[
                start_pos : end_pos if end_pos > 0 else None
            ]

            if "'" in encoded_filename:
                parts = encoded_filename.split("'", 2)
                if len(parts) == 3:
                    _, _, encoded_part = parts
                    filename = unquote(encoded_part.strip())
        except Exception:
            filename = None

    if not filename and "filename=" in content_disposition:
        try:
            start_pos = content_disposition.find("filename=") + len("filename=")
            # Handle quoted filenames
            if content_disposition[start_pos : start_pos + 1] == '"':
                end_pos = content_disposition.find('"', start_pos + 1)
                if end_pos > start_pos:
                    filename = content_disposition[start_pos + 1 : end_pos]
            else:
                # Non-quoted filename - read until semicolon or end
                end_pos = content_disposition.find(";", start_pos)
                filename = content_disposition[
                    start_pos : end_pos if end_pos > 0 else None
                ]

            filename = unquote(filename.strip())
        except Exception:
            filename = None

    # Fallback to URL if filename wasn't found
    if not filename:
        path_part = urlparse(url).path
        filename = unquote(path_part.split("/")[-1])

        if not filename or filename.startswith("?"):
            domain = urlparse(url).netloc.split(":")[0]  # Remove port if present
            filename = domain or "download"

    # Sanitize filename to avoid OS errors (replace invalid chars)
    invalid_chars = ["<", ">", ":", '"', "/", "\\", "|", "?", "*"]
    for char in invalid_chars:
        filename = filename.replace(char, "_")

    # If filename is still empty after all processing, use a default name
    if not filename:
        filename = "download"

    # Trim to reasonable filename length
    if len(filename.encode("utf-8")) > MAX_FILENAME_LENGTH:
        name, ext = path.splitext(filename)
        max_name_len = MAX_FILENAME_LENGTH - len(ext.encode("utf-8"))
        filename = (
            name.encode("utf-8")[:max_name_len].decode("utf-8", errors="ignore") + ext
        )

    if file_path:
        if await aio_os.path.isdir(file_path):
            return path.join(file_path, filename)
        return file_path

    return filename


async def create_segment_table(
    url: str,
    file_path: str,
    segments: int,
    size: Size,
    etag: str,
    etag_validation: bool,
) -> Dict:
    """Create a segment table for multi-segment download."""
    progress_file = file_path + ".json"
    overwrite = True

    if await aio_os.path.exists(progress_file):
        async with fopen(progress_file, "r") as f:
            progress = json.loads(await f.read())
            if not etag_validation or (
                progress["etag"]
                and (progress["url"] == url and progress["etag"] == etag)
            ):
                segments = progress["segments"]
                overwrite = False

    async with fopen(progress_file, "w") as f:
        await f.write(
            json.dumps(
                {"url": url, "etag": etag, "segments": segments},
                indent=4,
            )
        )

    dic = {"url": url, "segments": segments, "overwrite": overwrite}
    partition_size, add_bytes = divmod(size.value, segments)

    for segment in range(segments):
        start = size.start + partition_size * segment
        end = (
            size.start + partition_size * (segment + 1) - 1
        )  # since range is inclusive[0-99]

        if segment == segments - 1:
            end += add_bytes

        dic[segment] = {
            "segment_size": Size(start, end),
            "segment_path": f"{file_path}.{segment}",
        }

    return dic


async def combine_files(file_path: str, segments: int) -> None:
    """Combine the downloaded file segments into a single file."""
    async with fopen(file_path, "wb") as dest:
        for segment in range(segments):
            segment_file = f"{file_path}.{segment}"
            async with fopen(segment_file, "rb") as src:
                while True:
                    chunk = await src.read(CHUNKSIZE)
                    if chunk:
                        await dest.write(chunk)
                    else:
                        break

            await aio_os.remove(segment_file)

    progress_file = f"{file_path}.json"
    await aio_os.remove(progress_file)


def default_logger(name: str) -> logging.Logger:
    """Creates a default debugging logger."""
    logger = logging.getLogger(name)
    logger.setLevel(logging.WARN)
    handler = logging.FileHandler("pypdl.log", mode="a", delay=True)
    handler.setFormatter(
        logging.Formatter(
            "(%(name)s)  %(asctime)s - %(levelname)s: %(message)s",
            datefmt="%d-%m-%y %H:%M:%S",
        )
    )
    logger.addHandler(handler)
    return logger


def get_range(range_header: str, file_size: int) -> Tuple[int, int]:
    def parse_part(part: str) -> Optional[int]:
        return int(part) if part else None

    range_value = range_header.replace("bytes=", "")
    parts = range_value.split("-")
    if len(parts) != 2:
        raise TypeError("Invalid range format")

    start, end = map(parse_part, parts)

    if start is not None and end is not None:
        if start > end:
            raise TypeError("Invalid range, start is greater than end")
    else:
        if file_size == 0:
            raise TypeError("Invalid range, file size is 0")

        if end is not None:
            if end > file_size - 1:
                raise TypeError("Invalid range, end is greater than file size")
            start = file_size - end
            end = file_size - 1
        elif start is not None:
            if start > file_size - 1:
                raise TypeError("Invalid range, start is greater than file size")
            end = file_size - 1
        else:
            raise TypeError(f"Invalid range: {start}-{end}")

    return start, end



import asyncio
import time

import aiofiles
from aiohttp import ClientSession

MEGABYTE = 1048576


class Basicdown:
    """Base downloader class."""

    def __init__(self, session: ClientSession, speed_limit: float) -> None:
        self.session = session
        self.speed_limit = speed_limit * MEGABYTE
        self.curr = 0

    async def download(self, url: str, path: str, mode: str, **kwargs) -> None:
        """Download data in chunks."""
        speedlimit_time = time.time()
        speedlimit_size = 0
        async with self.session.get(url, **kwargs) as response:
            async with aiofiles.open(path, mode) as file:
                async for chunk in response.content.iter_chunked(MEGABYTE):
                    if self.speed_limit > 0:
                        now = time.time()
                        time_passed = now - speedlimit_time
                        if time_passed > 0.1:
                            curr_download = self.curr - speedlimit_size
                            if curr_download / time_passed >= self.speed_limit:
                                await asyncio.sleep(curr_download / self.speed_limit)
                            else:
                                speedlimit_time = now
                                speedlimit_size = self.curr

                    await file.write(chunk)
                    self.curr += len(chunk)


class Singledown(Basicdown):
    """Class for downloading the whole file in a single segment."""

    async def worker(self, url: str, file_path: str, **kwargs) -> None:
        await self.download(url, file_path, "wb", **kwargs)


class Multidown(Basicdown):
    """Class for downloading a specific segment of the file."""

    async def worker(self, segment_table: dict, id: int, **kwargs) -> None:
        url = segment_table["url"]
        overwrite = segment_table["overwrite"]
        segment_path = segment_table[id]["segment_path"]
        size = segment_table[id]["segment_size"]

        if await aiofiles.os.path.exists(segment_path):
            downloaded_size = await aiofiles.os.path.getsize(segment_path)
            if overwrite or downloaded_size > size.value:
                await aiofiles.os.remove(segment_path)
            else:
                self.curr = downloaded_size

        if kwargs.get("headers") is not None:
            kwargs["headers"] = kwargs["headers"].copy()

        if self.curr < size.value:
            start = size.start + self.curr
            kwargs.setdefault("headers", {}).update(
                {"range": f"bytes={start}-{size.end}"}
            )
            await self.download(url, segment_path, "ab", **kwargs)

        if self.curr != size.value:
            raise Exception(
                f"Incorrect segment size: expected {size} bytes, received {self.curr} bytes"
            )



import asyncio
import time
from collections import deque
from logging import Logger
from threading import Event
from typing import Callable, Union

import aiohttp

from . import utils
from .consumer import Consumer
from .producer import Producer


class Pypdl:
    def __init__(
        self,
        allow_reuse: bool = False,
        logger: Logger = utils.default_logger("Pypdl"),
        max_concurrent: int = 1,
    ):
        self._interrupt = Event()
        self._pool = utils.LoggingExecutor(logger, max_workers=1)
        self._loop = utils.TEventLoop()
        self._producer = None
        self._consumers = []
        self._producer_queue = None
        self._consumer_queue = None
        self._allow_reuse = allow_reuse
        self._logger = logger
        self._max_concurrent = max_concurrent
        self._future = None
        self._success = []
        self._failed = []

        self.size = None
        self.current_size = None
        self.remaining_size = None
        self.progress = None
        self.speed = None
        self.time_spent = None
        self.eta = None
        self.total_task = None
        self.completed_task = None
        self.task_progress = None
        self.completed = False

    @property
    def is_idle(self) -> bool:
        """Return True if download manager is idle, otherwise returns False"""
        return not self._loop.has_running_tasks()

    @property
    def logger(self) -> Union[Logger, None]:
        """Return logger if logger is in accessible state, otherwise returns None"""
        if self.is_idle:
            return self._logger
        return None

    @property
    def success(self) -> list:
        """Return list of successful downloads"""
        return self._success

    @property
    def failed(self) -> list:
        """Return list of failed downloads"""
        return self._failed

    def set_allow_reuse(self, allow_reuse) -> bool:
        """Returns True if allow_reuse was successfully set, otherwise returns False"""
        if self.is_idle:
            self._allow_reuse = allow_reuse
            return True
        return False

    def set_logger(self, logger) -> bool:
        """Returns True if logger was successfully set, otherwise returns False"""
        if self.is_idle:
            self._logger = logger
            return True
        return False

    def set_max_concurrent(self, max_concurrent) -> bool:
        """Returns True if max_concurrent was successfully set, otherwise returns False"""
        if self.is_idle:
            self._max_concurrent = max_concurrent
            return True
        return False

    def start(
        self,
        url: Union[Callable, str] = None,
        file_path: str = None,
        tasks: list = None,
        multisegment: bool = True,
        segments: int = 5,
        retries: int = 0,
        mirrors: Union[str, list, Callable] = None,
        overwrite: bool = True,
        speed_limit: float = 0,
        etag_validation: bool = True,
        block: bool = True,
        display: bool = True,
        clear_terminal: bool = True,
        **kwargs,
    ) -> Union[utils.EFuture, utils.AutoShutdownFuture, list]:
        """
        Start the download process.

        :param url: URL (str or callable) if `tasks` not provided.
        :param file_path: Local file path for the single download.
        :param tasks: List of tasks (dict), each with its own URL and other supported keys.
        :param multisegment: Whether to download in multiple segments.
        :param segments: Number of segments if multi-segmented.
        :param retries: Number of retries for failed downloads.
        :param mirrors: List of mirrors (callable/string) or callable/string.
        :param overwrite: Overwrite existing files if True.
        :param speed_limit: Limit download speed in MB/s if > 0.
        :param etag_validation: Validate server-provided ETag if True.
        :param block: If True, block until downloads finish.
        :param display: If True, display progress.
        :param clear_terminal: If True, clear terminal before displaying progress bar.
        :param kwargs: Addtional keyword arguments for aiohttp.
        :return: A future-like object if non-blocking, or a result list if blocking.
        :raises RuntimeError: If downloads are already in progress.
        :raises TypeError: If invalid parameters or task definitions are provided.
        """
        if not self.is_idle:
            tasks = asyncio.all_tasks(self._loop.loop)
            raise RuntimeError(f"Pypdl already running {tasks}")

        if tasks and (url or file_path):
            raise TypeError(
                "Either provide 'tasks' or a 'url' (with optional 'file_path'), but not both."
            )

        if tasks is None:
            if not (isinstance(url, str) or callable(url)):
                raise TypeError(
                    "Expected a 'url' (str or callable) when 'tasks' not provided."
                )
            tasks = [{"url": url, "file_path": file_path, "mirrors": mirrors}]

        self._reset()

        task_dict = {}
        self.total_task = 0
        _kwargs = {
            "timeout": aiohttp.ClientTimeout(sock_read=60),
            "raise_for_status": True,
        }
        _kwargs.update(kwargs)
        for i, task_kwargs in enumerate(tasks):
            task = utils.Task(
                multisegment,
                segments,
                retries,
                overwrite,
                speed_limit,
                etag_validation,
                **_kwargs,
            )
            task.set(**task_kwargs)
            task_dict[i] = task
            self.total_task += 1

        coro = self._download_tasks(task_dict, display, clear_terminal)

        self._future = utils.EFuture(
            asyncio.run_coroutine_threadsafe(coro, self._loop.get()),
            self._loop,
            self._interrupt,
        )

        while self.is_idle:
            self._logger.debug("waiting for download to start")
            time.sleep(0.1)

        if not self._allow_reuse:
            future = utils.AutoShutdownFuture(self._future, self._loop, self._pool)
        else:
            future = self._future

        if block:
            result = future.result()
            return result

        return future

    async def _download_tasks(self, tasks_dict, display, clear_terminal):
        self._logger.debug("Starting download tasks")
        start_time = time.time()
        coroutines = []
        self._producer_queue = asyncio.Queue(self.total_task)
        self._consumer_queue = asyncio.Queue(self.total_task)

        try:
            async with aiohttp.ClientSession() as session:
                self._producer = Producer(session, self._logger, tasks_dict)
                producer_task = asyncio.create_task(
                    self._producer.enqueue_tasks(
                        self._producer_queue, self._consumer_queue
                    )
                )
                coroutines.append(producer_task)
                await self._producer_queue.put(list(tasks_dict))

                for _id in range(self._max_concurrent):
                    consumer = Consumer(session, self._logger, _id)
                    consumer_task = asyncio.create_task(
                        consumer.process_tasks(
                            self._consumer_queue,
                            self._producer_queue,
                        )
                    )
                    coroutines.append(consumer_task)
                    self._consumers.append(consumer)

                self._logger.debug("Starting producer and consumer tasks")
                self._pool.submit(self._progress_monitor, display, clear_terminal)
                await utils.auto_cancel_gather(*coroutines)
                await asyncio.sleep(0.5)
        except utils.MainThreadException as e:
            self._interrupt.set()
            self._logger.exception(e)
            raise e

        self.time_spent = time.time() - start_time
        if display:
            print(f"Time elapsed: {utils.seconds_to_hms(self.time_spent)}")

        return self._success

    def stop(self) -> None:
        """Stop the download manager."""
        self._logger.debug("stop called")
        if self.is_idle or self.completed:
            self._logger.debug("Task not running, nothing to stop")
            return None
        self._future._stop()

    def shutdown(self) -> None:
        """Shutdown the download manager."""
        self.stop()
        self._loop.stop()
        self._pool.shutdown()
        self._logger.debug("Shutdown download manager")

    def _reset(self):
        self._interrupt.clear()
        self._consumers.clear()
        self._producer = None

        self.size = None
        self.current_size = None
        self.remaining_size = None
        self.progress = None
        self.speed = None
        self.time_spent = None
        self.eta = None
        self.total_task = None
        self.completed_task = None
        self.task_progress = None
        self.completed = False
        self._success.clear()
        self._failed.clear()

    def _progress_monitor(self, display, clear_terminal):
        self._logger.debug("Starting progress monitor")
        interval = 0.5
        recent_queue = deque(maxlen=12)
        with utils.ScreenCleaner(display, clear_terminal):
            while not self.completed and not self._interrupt.is_set():
                self._calc_values(recent_queue, interval)
                if display:
                    self._display()
                time.sleep(interval)
        self._logger.debug("exiting progress monitor")

    def _calc_values(self, recent_queue, interval):
        self.size = self._producer.size
        self.current_size = sum(consumer.size for consumer in self._consumers)
        self._success = sum([consumer.success for consumer in self._consumers], [])
        self._failed = self._producer.failed
        self.completed_task = len(self._success) + len(self._failed)
        self.task_progress = int((self.completed_task / self.total_task) * 100)

        # Speed calculation
        recent_queue.append(self.current_size)
        non_zero_list = [utils.to_mb(value) for value in recent_queue if value]
        if len(non_zero_list) < 1:
            self.speed = 0
        elif len(non_zero_list) == 1:
            self.speed = non_zero_list[0] / interval
        else:
            diff = [b - a for a, b in zip(non_zero_list, non_zero_list[1:])]
            self.speed = (sum(diff) / len(diff)) / interval

        if self.size:
            self.progress = int((self.current_size / self.size) * 100)
            self.remaining_size = utils.to_mb(self.size - self.current_size)

            if self.speed:
                self.eta = self.remaining_size / self.speed
            else:
                self.eta = -1

        if self.completed_task is self.total_task:
            future = asyncio.run_coroutine_threadsafe(
                self._completed(), self._loop.get()
            )
            future.result()

    async def _completed(self):
        self._logger.debug("All downloads completed")
        await self._producer_queue.put(None)
        for _ in range(self._max_concurrent):
            await self._consumer_queue.put(None)
        self.completed = True

    def _display(self):
        utils.cursor_up()
        whitespace = " "
        if self.size:
            progress_bar = f"[{'█' * self.progress}{'·' * (100 - self.progress)}] {self.progress}% \n"

            if self.total_task > 1:
                info1 = f"Total Downloads: {self.completed_task}/{self.total_task}, "
            else:
                info1 = ""

            info2 = f"Size: {utils.to_mb(self.size):.2f} MB, Speed: {self.speed:.2f} MB/s, ETA: {utils.seconds_to_hms(self.eta)}"
            print(progress_bar + info1 + info2 + whitespace * 35)
        else:
            if self.total_task > 1:
                download_stats = f"[{'█' * self.task_progress}{'·' * (100 - self.task_progress)}] {self.task_progress}% \n"
            else:
                download_stats = f"Downloading... {whitespace * 95}\n"

            info = f"Total Downloads: {self.completed_task}/{self.total_task}, Downloaded Size: {utils.to_mb(self.current_size):.2f} MB, Speed: {self.speed:.2f} MB/s"

            print(download_stats + info + whitespace * 35)



import asyncio
from threading import Lock

from .utils import Size, check_main_thread_exception, get_filepath, get_range, get_url


class Producer:
    def __init__(self, session, logger, tasks):
        self._size = None
        self._failed = []
        self._lock = Lock()
        self._logger = logger
        self._session = session
        self._tasks = tasks
        self._size_avail = True

    @property
    def size(self):
        if self._size_avail:
            return self._size

    @property
    def failed(self):
        with self._lock:
            return self._failed.copy()

    def add_failed(self, item):
        with self._lock:
            self._failed.append(item)

    async def enqueue_tasks(self, in_queue, out_queue):
        self._logger.debug("Producer started")
        while True:
            batch = await in_queue.get()

            if batch is None:
                break

            total_size = 0

            for _id in batch:
                task = self._tasks[_id]
                if task.tries > 0:
                    task.tries -= 1
                    try:
                        (
                            url,
                            file_path,
                            multisegment,
                            etag,
                            size,
                            kwargs,
                        ) = await self._fetch_task_info(
                            task.url, task.file_path, task.multisegment, **task.kwargs
                        )
                    except asyncio.CancelledError:
                        raise
                    except Exception as e:
                        check_main_thread_exception(e)
                        self._logger.debug(
                            f"Failed to get header for {task}, skipping task"
                        )
                        self._logger.exception(e)
                        await asyncio.sleep(3)
                        new_url = (
                            task.mirrors.pop(0) if task.mirrors else task.default_url
                        )
                        task.url = new_url
                        await in_queue.put([_id])
                        continue

                    if size.value == 0:
                        self._logger.debug("Size is Unavailable, setting size to None")
                        self._size_avail = False

                    total_size -= task.size.value
                    total_size += size.value
                    task.size = size
                    await out_queue.put(
                        (
                            _id,
                            (
                                url,
                                file_path,
                                multisegment,
                                etag,
                                size,
                                task.segments,
                                task.overwrite,
                                task.speed_limit,
                                task.etag_validation,
                                kwargs,
                            ),
                        )
                    )
                else:
                    self.add_failed(task.url)

            if self._size is None:
                self._size = total_size
            else:
                self._size += total_size
        self._logger.debug("Producer exited")

    async def _fetch_task_info(self, url, file_path, multisegment, **kwargs):
        url = await get_url(url)

        user_headers = kwargs.get("headers", {})
        range_header = None

        if user_headers:
            _user_headers = user_headers.copy()
            for key, value in user_headers.items():
                if key.lower() == "range":
                    range_header = value
                    self._logger.debug("Range header found %s", range_header)
                    del _user_headers[key]
            kwargs["headers"] = _user_headers

        header = await self._fetch_header(url, **kwargs)
        file_path = await get_filepath(url, header, file_path)
        if file_size := int(header.get("content-length", 0)):
            self._logger.debug("File size acquired from header")

        if range_header:
            start, end = get_range(range_header, file_size)
        else:
            start = 0
            end = file_size - 1

        size = Size(start, end)
        etag = header.get("etag", "")
        if etag != "":
            self._logger.debug("ETag acquired from header")
            etag = etag.strip('"')

        if size.value < 1 or not header.get("accept-ranges"):
            self._logger.debug("Single segment mode, accept-ranges or size not found")
            kwargs["headers"] = user_headers
            multisegment = False

        return url, file_path, multisegment, etag, size, kwargs

    async def _fetch_header(self, url, **kwargs):
        try:
            async with self._session.head(url, **kwargs) as response:
                if response.status < 400:
                    self._logger.debug("Header acquired from HEAD request")
                    return response.headers
        except Exception:
            pass

        async with self._session.get(url, **kwargs) as response:
            if response.status < 400:
                self._logger.debug("Header acquired from GET request")
                return response.headers



from setuptools import setup, find_packages

DESCRIPTION = "A concurrent pure python download manager"
with open("README.md", "r") as f:
    LONG_DESCRIPTION = f.read()

with open("pypdl/__init__.py", "r") as f:
    VERSION = f.readline().split("=")[1].strip().strip("\"'")

setup(
    name="pypdl",
    version=VERSION,
    author="mjishnu",
    author_email="<mjishnu@proton.me>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    url="https://github.com/mjishnu/pypdl",
    license="MIT",
    classifiers={
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    },
    packages=find_packages(),
    install_requires=["aiohttp", "aiofiles"],
    keywords=[
        "python",
        "downloader",
        "multi-threaded-downloader",
        "concurrent-downloader",
        "parallel-downloader",
        "async-downloader",
        "asyncronous-downloader",
        "download-manager",
        "fast-downloader",
        "download-accelerator",
        "download-optimizer",
        "download-utility",
        "download-tool",
        "download-automation",
    ],
    python_requires=">=3.8",
)
